# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.utils.py3 import httplib

from libcloud.dns.types import RecordType
from libcloud.dns.types import ZoneDoesNotExistError
from libcloud.dns.types import RecordDoesNotExistError
from libcloud.dns.drivers.pointdns import PointDNSDriver
from libcloud.dns.drivers.pointdns import PointDNSException

from libcloud.test import MockHttp
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.test.secrets import DNS_PARAMS_POINTDNS


class PointDNSTests(unittest.TestCase):
    def setUp(self):
        PointDNSDriver.connectionCls.conn_class = PointDNSMockHttp
        PointDNSMockHttp.type = None
        self.driver = PointDNSDriver(*DNS_PARAMS_POINTDNS)

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' %
                            (key))

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 10)
        self.assertTrue(RecordType.A in record_types)
        self.assertTrue(RecordType.AAAA in record_types)
        self.assertTrue(RecordType.ALIAS in record_types)
        self.assertTrue(RecordType.CNAME in record_types)
        self.assertTrue(RecordType.MX in record_types)
        self.assertTrue(RecordType.NS in record_types)
        self.assertTrue(RecordType.PTR in record_types)
        self.assertTrue(RecordType.SRV in record_types)
        self.assertTrue(RecordType.SSHFP in record_types)
        self.assertTrue(RecordType.TXT in record_types)

    def test_list_zones_success(self):
        PointDNSMockHttp.type = 'GET'
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)

        zone1 = zones[0]
        self.assertEqual(zone1.id, '1')
        self.assertEqual(zone1.type, 'master')
        self.assertEqual(zone1.domain, 'example.com')
        self.assertEqual(zone1.ttl, 3600)
        self.assertHasKeys(zone1.extra, ['group', 'user-id'])

        zone2 = zones[1]
        self.assertEqual(zone2.id, '2')
        self.assertEqual(zone2.type, 'master')
        self.assertEqual(zone2.domain, 'example2.com')
        self.assertEqual(zone2.ttl, 3600)
        self.assertHasKeys(zone2.extra, ['group', 'user-id'])

    def test_list_records_success(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 2)

        record1 = records[0]
        self.assertEqual(record1.id, '141')
        self.assertEqual(record1.name, 'site.example.com')
        self.assertEqual(record1.type, RecordType.A)
        self.assertEqual(record1.data, '1.2.3.4')
        self.assertHasKeys(record1.extra, ['ttl', 'zone_id', 'aux'])

        record2 = records[1]
        self.assertEqual(record2.id, '150')
        self.assertEqual(record2.name, 'site.example1.com')
        self.assertEqual(record2.type, RecordType.A)
        self.assertEqual(record2.data, '1.2.3.6')
        self.assertHasKeys(record2.extra, ['ttl', 'zone_id', 'aux'])

    def test_get_zone_success(self):
        PointDNSMockHttp.type = 'GET'
        zone1 = self.driver.get_zone(zone_id='1')
        self.assertEqual(zone1.id, '1')
        self.assertEqual(zone1.type, 'master')
        self.assertEqual(zone1.domain, 'example.com')
        self.assertEqual(zone1.ttl, 3600)
        self.assertHasKeys(zone1.extra, ['group', 'user-id'])

    def test_get_zone_zone_not_exists(self):
        PointDNSMockHttp.type = 'GET_ZONE_NOT_EXIST'
        try:
            self.driver.get_zone(zone_id='1')
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_get_record_success(self):
        PointDNSMockHttp.type = 'GET'
        record = self.driver.get_record(zone_id='1',
                                        record_id='141')
        self.assertEqual(record.id, '141')
        self.assertEqual(record.name, 'site.example.com')
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, '1.2.3.4')
        self.assertHasKeys(record.extra, ['ttl', 'zone_id', 'aux'])

    def test_get_record_record_not_exists(self):
        PointDNSMockHttp.type = 'GET_RECORD_NOT_EXIST'
        try:
            self.driver.get_record(zone_id='1',
                                   record_id='141')
        except RecordDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_create_zone_success(self):
        PointDNSMockHttp.type = 'CREATE'
        zone = self.driver.create_zone(domain='example.com')
        self.assertEqual(zone.id, '2')
        self.assertEqual(zone.domain, 'example.com')
        self.assertEqual(zone.ttl, 3600)
        self.assertEqual(zone.type, 'master')
        self.assertHasKeys(zone.extra, ['group', 'user-id'])

    def test_create_zone_with_error(self):
        PointDNSMockHttp.type = 'CREATE_ZONE_WITH_ERROR'
        try:
            self.driver.create_zone(domain='example.com')
        except PointDNSException:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_create_record_success(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'CREATE'
        record = self.driver.create_record(name='site.example.com', zone=zone,
                                           type=RecordType.A,
                                           data='1.2.3.4')
        self.assertEqual(record.id, '143')
        self.assertEqual(record.name, 'site.example.com')
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, '1.2.3.4')
        self.assertHasKeys(record.extra, ['ttl', 'zone_id', 'aux'])

    def test_create_record_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'CREATE_WITH_ERROR'
        try:
            self.driver.create_record(name='site.example.com',
                                      zone=zone, type=RecordType.A,
                                      data='1.2.3.4')
        except PointDNSException:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_update_zone_success(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'ZONE_UPDATE'
        extra = {'user-id': 6}
        _zone = self.driver.update_zone(zone, zone.domain, zone.ttl,
                                        extra=extra)
        self.assertEqual(_zone.extra.get('user-id'), 6)

    def test_update_zone_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'UPDATE_ZONE_WITH_ERROR'
        extra = {'user-id': 6}
        try:
            self.driver.update_zone(zone, zone.domain, zone.ttl, extra=extra)
        except PointDNSException:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_update_record_success(self):
        PointDNSMockHttp.type = 'GET'
        record = self.driver.get_record(zone_id='1',
                                        record_id='141')
        PointDNSMockHttp.type = 'UPDATE'
        extra = {'ttl': 4500}
        record1 = self.driver.update_record(record=record, name='updated.com',
                                            type=RecordType.A, data='1.2.3.5',
                                            extra=extra)
        self.assertEqual(record.data, '1.2.3.4')
        self.assertEqual(record.extra.get('ttl'), 3600)
        self.assertEqual(record1.data, '1.2.3.5')
        self.assertEqual(record1.extra.get('ttl'), 4500)

    def test_update_record_with_error(self):
        PointDNSMockHttp.type = 'GET'
        record = self.driver.get_record(zone_id='1',
                                        record_id='141')
        PointDNSMockHttp.type = 'UPDATE_RECORD_WITH_ERROR'
        extra = {'ttl': 4500}
        try:
            self.driver.update_record(record=record, name='updated.com',
                                      type=RecordType.A, data='1.2.3.5',
                                      extra=extra)
        except PointDNSException:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_delete_zone_success(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'DELETE'
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_zone_not_exists(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'DELETE_ZONE_NOT_EXIST'
        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_delete_record_success(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 2)
        record = records[1]
        PointDNSMockHttp.type = 'DELETE'
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_record_not_exists(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 2)
        record = records[1]
        PointDNSMockHttp.type = 'DELETE_RECORD_NOT_EXIST'
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_ex_list_redirects_success(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'LIST'
        redirects = self.driver.ex_list_redirects(zone)
        self.assertEqual(len(redirects), 2)

        redirect1 = redirects[0]
        self.assertEqual(redirect1.id, '36843229')
        self.assertEqual(redirect1.name, 'redirect2.domain1.com.')
        self.assertEqual(redirect1.type, '302')
        self.assertEqual(redirect1.data, 'http://other.com')
        self.assertIsNone(redirect1.iframe)
        self.assertEqual(redirect1.query, False)
        self.assertEqual(zone, redirect1.zone)

        redirect2 = redirects[1]
        self.assertEqual(redirect2.id, '36843497')
        self.assertEqual(redirect2.name, 'redirect1.domain1.com.')
        self.assertEqual(redirect2.type, '302')
        self.assertEqual(redirect2.data, 'http://someother.com')
        self.assertIsNone(redirect2.iframe)
        self.assertEqual(redirect2.query, False)
        self.assertEqual(zone, redirect1.zone)

    def test_ex_list_mail_redirects(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'LIST'
        mail_redirects = self.driver.ex_list_mail_redirects(zone)
        self.assertEqual(len(mail_redirects), 2)

        mail_redirect1 = mail_redirects[0]
        self.assertEqual(mail_redirect1.id, '5')
        self.assertEqual(mail_redirect1.source, 'admin')
        self.assertEqual(mail_redirect1.destination, 'user@example-site.com')
        self.assertEqual(zone, mail_redirect1.zone)

        mail_redirect2 = mail_redirects[1]
        self.assertEqual(mail_redirect2.id, '7')
        self.assertEqual(mail_redirect2.source, 'new_admin')
        self.assertEqual(mail_redirect2.destination,
                         'second.user@example-site.com')
        self.assertEqual(zone, mail_redirect2.zone)

    def test_ex_create_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'CREATE'
        redirect = self.driver.ex_create_redirect('http://other.com',
                                                  'redirect2', '302', zone,
                                                  iframe='An Iframe',
                                                  query=True)
        self.assertEqual(redirect.id, '36843229')
        self.assertEqual(redirect.name, 'redirect2.domain1.com.')
        self.assertEqual(redirect.type, '302')
        self.assertEqual(redirect.data, 'http://other.com')
        self.assertEqual(redirect.iframe, 'An Iframe')
        self.assertEqual(redirect.query, True)
        self.assertEqual(zone.id, redirect.zone.id)

    def test_ex_create_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'CREATE_WITH_ERROR'
        try:
            self.driver.ex_create_redirect('http://other.com', 'redirect2',
                                           '302', zone, iframe='An Iframe',
                                           query=True)
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_create_mail_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'CREATE'
        mail_redirect = self.driver.ex_create_mail_redirect(
            'user@example-site.com', 'admin', zone)
        self.assertEqual(mail_redirect.id, '5')
        self.assertEqual(mail_redirect.source, 'admin')
        self.assertEqual(mail_redirect.destination, 'user@example-site.com')
        self.assertEqual(zone.id, mail_redirect.zone.id)

    def test_ex_create_mail_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'CREATE_WITH_ERROR'
        try:
            self.driver.ex_create_mail_redirect('user@example-site.com',
                                                'admin', zone)
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_get_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        redirect = self.driver.ex_get_redirect(zone.id, '36843229')
        self.assertEqual(redirect.id, '36843229')
        self.assertEqual(redirect.name, 'redirect2.domain1.com.')
        self.assertEqual(redirect.type, '302')
        self.assertEqual(redirect.data, 'http://other.com')
        self.assertIsNone(redirect.iframe)
        self.assertEqual(redirect.query, False)
        self.assertEqual(zone.id, redirect.zone.id)

    def test_ex_get_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'GET_WITH_ERROR'
        try:
            self.driver.ex_get_redirect(zone.id, '36843229')
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_get_redirect_not_found(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'GET_NOT_FOUND'
        try:
            self.driver.ex_get_redirect(zone.id, '36843229')
        except PointDNSException as e:
            self.assertEqual(e.http_code, httplib.NOT_FOUND)
            self.assertEqual(e.value, "Couldn't found redirect")
        else:
            self.fail('Exception was not thrown')

    def test_ex_get_mail_redirects(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        mail_redirect = self.driver.ex_get_mail_redirects(zone.id, '5')
        self.assertEqual(mail_redirect.id, '5')
        self.assertEqual(mail_redirect.source, 'admin')
        self.assertEqual(mail_redirect.destination, 'user@example-site.com')
        self.assertEqual(zone.id, mail_redirect.zone.id)

    def test_ex_get_mail_redirects_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        PointDNSMockHttp.type = 'GET_WITH_ERROR'
        try:
            self.driver.ex_get_mail_redirects(zone.id, '5')
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_update_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        redirect = self.driver.ex_get_redirect(zone.id, '36843229')
        PointDNSMockHttp.type = 'UPDATE'
        _redirect = self.driver.ex_update_redirect(
            redirect, 'http://updatedother.com', 'redirect3', '302')
        self.assertEqual(_redirect.id, '36843229')
        self.assertEqual(_redirect.name, 'redirect3.domain1.com.')
        self.assertEqual(_redirect.type, '302')
        self.assertEqual(_redirect.data, 'http://updatedother.com')
        self.assertIsNone(_redirect.iframe)
        self.assertEqual(_redirect.query, False)
        self.assertEqual(zone.id, _redirect.zone.id)

    def test_ex_update_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        redirect = self.driver.ex_get_redirect(zone.id, '36843229')
        PointDNSMockHttp.type = 'UPDATE_WITH_ERROR'
        try:
            self.driver.ex_update_redirect(
                redirect, 'http://updatedother.com', 'redirect3', '302')
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_update_mail_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        mailredirect = self.driver.ex_get_mail_redirects(zone.id, '5')
        PointDNSMockHttp.type = 'UPDATE'
        _mailredirect = self.driver.ex_update_mail_redirect(
            mailredirect, 'new_user@example-site.com', 'new_admin')
        self.assertEqual(_mailredirect.id, '5')
        self.assertEqual(_mailredirect.source, 'new_admin')
        self.assertEqual(_mailredirect.destination,
                         'new_user@example-site.com')
        self.assertEqual(zone.id, _mailredirect.zone.id)

    def test_ex_update_mail_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        mailredirect = self.driver.ex_get_mail_redirects(zone.id, '5')
        PointDNSMockHttp.type = 'UPDATE_WITH_ERROR'
        try:
            self.driver.ex_update_mail_redirect(
                mailredirect, 'new_user@example-site.com', 'new_admin')
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_delete_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        redirect = self.driver.ex_get_redirect(zone.id, '36843229')
        PointDNSMockHttp.type = 'DELETE'
        status = self.driver.ex_delete_redirect(redirect)
        self.assertTrue(status)

    def test_ex_delete_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        redirect = self.driver.ex_get_redirect(zone.id, '36843229')
        PointDNSMockHttp.type = 'DELETE_WITH_ERROR'
        try:
            self.driver.ex_delete_redirect(redirect)
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_delete_redirect_not_found(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        redirect = self.driver.ex_get_redirect(zone.id, '36843229')
        PointDNSMockHttp.type = 'DELETE_NOT_FOUND'
        try:
            self.driver.ex_delete_redirect(redirect)
        except PointDNSException as e:
            self.assertEqual(e.http_code, httplib.NOT_FOUND)
            self.assertEqual(e.value, "Couldn't found redirect")
        else:
            self.fail('Exception was not thrown')

    def test_ex_delete_mail_redirect(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        mailredirect = self.driver.ex_get_mail_redirects(zone.id, '5')
        PointDNSMockHttp.type = 'DELETE'
        status = self.driver.ex_delete_mail_redirect(mailredirect)
        self.assertTrue(status)

    def test_ex_delete_mail_redirect_with_error(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        mailredirect = self.driver.ex_get_mail_redirects(zone.id, '5')
        PointDNSMockHttp.type = 'DELETE_WITH_ERROR'
        try:
            self.driver.ex_delete_mail_redirect(mailredirect)
        except PointDNSException as e:
            # The API actually responds with httplib.UNPROCESSABLE_ENTITY code,
            # but httplib.responses doesn't have it.
            self.assertEqual(e.http_code, httplib.METHOD_NOT_ALLOWED)
        else:
            self.fail('Exception was not thrown')

    def test_ex_delete_mail_redirect_not_found(self):
        PointDNSMockHttp.type = 'GET'
        zone = self.driver.list_zones()[0]
        mailredirect = self.driver.ex_get_mail_redirects(zone.id, '5')
        PointDNSMockHttp.type = 'DELETE_NOT_FOUND'
        try:
            self.driver.ex_delete_mail_redirect(mailredirect)
        except PointDNSException as e:
            self.assertEqual(e.http_code, httplib.NOT_FOUND)
            self.assertEqual(e.value, "Couldn't found mail redirect")
        else:
            self.fail('Exception was not thrown')


class PointDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures('pointdns')

    def _zones_GET(self, method, url, body, headers):
        body = self.fixtures.load('_zones_GET.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_CREATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_CREATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_CREATE_ZONE_WITH_ERROR(self, method, url, body, headers):
        body = self.fixtures.load('error.json')
        return (httplib.PAYMENT_REQUIRED, body, {},
                httplib.responses[httplib.PAYMENT_REQUIRED])

    def _zones_1_GET(self, method, url, body, headers):
        body = self.fixtures.load('_zones_GET_1.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_ZONE_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_ZONE_UPDATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_UPDATE_ZONE_WITH_ERROR(self, method, url, body, headers):
        body = self.fixtures.load('error.json')
        return (httplib.PAYMENT_REQUIRED, body, {},
                httplib.responses[httplib.PAYMENT_REQUIRED])

    def _zones_1_GET_ZONE_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_example_com_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_example_com_UPDATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_DELETE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_DELETE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_DELETE_ZONE_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_records_CREATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_example_com_records_CREATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_records_CREATE_WITH_ERROR(self, method, url, body, headers):
        body = self.fixtures.load('error.json')
        return (httplib.PAYMENT_REQUIRED, body, {},
                httplib.responses[httplib.PAYMENT_REQUIRED])

    def _zones_1_records_GET(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_records_GET.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_records_141_GET_RECORD_NOT_EXIST(self, method, url, body,
                                                  headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_records_141_GET(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_records_141_GET.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_records_141_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_records_141_UPDATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_records_141_UPDATE_RECORD_WITH_ERROR(self, method, url, body,
                                                      headers):
        body = self.fixtures.load('error.json')
        return (httplib.PAYMENT_REQUIRED, body, {},
                httplib.responses[httplib.PAYMENT_REQUIRED])

    def _zones_1_records_150_DELETE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_records_150_DELETE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_records_150_DELETE_RECORD_NOT_EXIST(self, method, url, body,
                                                     headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_redirects_LIST(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_redirects_LIST.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_mail_redirects_LIST(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_mail_redirects_LIST.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_redirects_CREATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_redirects_CREATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_redirects_CREATE_WITH_ERROR(self, method, url, body, headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_mail_redirects_CREATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_mail_redirects_CREATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_mail_redirects_CREATE_WITH_ERROR(self, method, url, body,
                                                  headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_redirects_36843229_GET_WITH_ERROR(self, method, url, body,
                                                   headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_redirects_36843229_GET(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_redirects_GET.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_redirects_36843229_GET_NOT_FOUND(self, method, url, body,
                                                  headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_mail_redirects_5_GET(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_mail_redirects_GET.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_mail_redirects_5_GET_WITH_ERROR(self, method, url, body,
                                                 headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_redirects_36843229_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_redirects_UPDATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_redirects_36843229_UPDATE_WITH_ERROR(self, method, url, body,
                                                      headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_mail_redirects_5_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_mail_redirects_UPDATE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_mail_redirects_5_UPDATE_WITH_ERROR(self, method, url, body,
                                                    headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_redirects_36843229_DELETE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_redirects_DELETE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_mail_redirects_5_DELETE(self, method, url, body, headers):
        body = self.fixtures.load('_zones_1_mail_redirects_DELETE.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1_redirects_36843229_DELETE_WITH_ERROR(self, method, url, body,
                                                      headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_mail_redirects_5_DELETE_WITH_ERROR(self, method, url, body,
                                                    headers):
        body = self.fixtures.load('redirect_error.json')
        return (httplib.METHOD_NOT_ALLOWED, body, {},
                httplib.responses[httplib.METHOD_NOT_ALLOWED])

    def _zones_1_redirects_36843229_DELETE_NOT_FOUND(self, method, url, body,
                                                     headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_mail_redirects_5_DELETE_NOT_FOUND(self, method, url, body,
                                                   headers):
        body = self.fixtures.load('not_found.json')
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from libcloud.utils.py3 import httplib

try:
    import simplejson as json
except ImportError:
    import json

from libcloud.utils.misc import reverse_dict
from libcloud.common.types import LibcloudError
from libcloud.common.gogrid import GoGridConnection, GoGridResponse,\
    BaseGoGridDriver
from libcloud.loadbalancer.base import LoadBalancer, Member, Driver, Algorithm
from libcloud.loadbalancer.base import DEFAULT_ALGORITHM
from libcloud.loadbalancer.types import State, LibcloudLBImmutableError


class GoGridLBResponse(GoGridResponse):
    def success(self):
        if self.status == httplib.INTERNAL_SERVER_ERROR:
            # Hack, but at least this error message is more useful than
            # "unexpected server error"
            body = json.loads(self.body)
            if body['method'] == '/grid/loadbalancer/add' and \
                len(body['list']) >= 1 and \
                body['list'][0]['message'].find(
                    'unexpected server error') != -1:
                raise LibcloudError(
                    value='You mostly likely tried to add a member with an IP'
                          ' address not assigned to your account', driver=self)
        return super(GoGridLBResponse, self).success()


class GoGridLBConnection(GoGridConnection):
    """
    Connection class for the GoGrid load-balancer driver.
    """
    responseCls = GoGridLBResponse


class GoGridLBDriver(BaseGoGridDriver, Driver):
    connectionCls = GoGridLBConnection
    api_name = 'gogrid_lb'
    name = 'GoGrid LB'
    website = 'http://www.gogrid.com/'

    LB_STATE_MAP = {'On': State.RUNNING,
                    'Unknown': State.UNKNOWN}
    _VALUE_TO_ALGORITHM_MAP = {
        'round robin': Algorithm.ROUND_ROBIN,
        'least connect': Algorithm.LEAST_CONNECTIONS
    }
    _ALGORITHM_TO_VALUE_MAP = reverse_dict(_VALUE_TO_ALGORITHM_MAP)

    def __init__(self, *args, **kwargs):
        """
        @inherits: :class:`Driver.__init__`
        """
        super(GoGridLBDriver, self).__init__(*args, **kwargs)

    def list_protocols(self):
        # GoGrid only supports http
        return ['http']

    def list_balancers(self):
        return self._to_balancers(
            self.connection.request('/api/grid/loadbalancer/list').object)

    def ex_create_balancer_nowait(self, name, members, protocol='http',
                                  port=80, algorithm=DEFAULT_ALGORITHM):
        """
        @inherits: :class:`Driver.create_balancer`
        """
        algorithm = self._algorithm_to_value(algorithm)

        params = {'name': name,
                  'loadbalancer.type': algorithm,
                  'virtualip.ip': self._get_first_ip(),
                  'virtualip.port': port}
        params.update(self._members_to_params(members))

        resp = self.connection.request('/api/grid/loadbalancer/add',
                                       method='GET',
                                       params=params)
        return self._to_balancers(resp.object)[0]

    def create_balancer(self, name, members, protocol='http', port=80,
                        algorithm=DEFAULT_ALGORITHM):
        balancer = self.ex_create_balancer_nowait(name, members, protocol,
                                                  port, algorithm)

        timeout = 60 * 20
        waittime = 0
        interval = 2 * 15

        if balancer.id is not None:
            return balancer
        else:
            while waittime < timeout:
                balancers = self.list_balancers()

                for i in balancers:
                    if i.name == balancer.name and i.id is not None:
                        return i

                waittime += interval
                time.sleep(interval)

        raise Exception('Failed to get id')

    def destroy_balancer(self, balancer):
        try:
            resp = self.connection.request(
                '/api/grid/loadbalancer/delete', method='POST',
                params={'id': balancer.id})
        except Exception as e:
            if "Update request for LoadBalancer" in str(e):
                raise LibcloudLBImmutableError(
                    "Cannot delete immutable object", GoGridLBDriver)
            else:
                raise

        return resp.status == 200

    def get_balancer(self, **kwargs):
        params = {}

        try:
            params['name'] = kwargs['ex_balancer_name']
        except KeyError:
            balancer_id = kwargs['balancer_id']
            params['id'] = balancer_id

        resp = self.connection.request('/api/grid/loadbalancer/get',
                                       params=params)

        return self._to_balancers(resp.object)[0]

    def balancer_attach_member(self, balancer, member):
        members = self.balancer_list_members(balancer)
        members.append(member)

        params = {"id": balancer.id}

        params.update(self._members_to_params(members))

        resp = self._update_balancer(params)
        return [m for m in
                self._to_members(resp.object["list"][0]["realiplist"],
                                 balancer)
                if m.ip == member.ip][0]

    def balancer_detach_member(self, balancer, member):
        members = self.balancer_list_members(balancer)

        remaining_members = [n for n in members if n.id != member.id]

        params = {"id": balancer.id}
        params.update(self._members_to_params(remaining_members))

        resp = self._update_balancer(params)

        return resp.status == 200

    def balancer_list_members(self, balancer):
        resp = self.connection.request('/api/grid/loadbalancer/get',
                                       params={'id': balancer.id})
        return self._to_members(resp.object["list"][0]["realiplist"], balancer)

    def _update_balancer(self, params):
        try:
            return self.connection.request('/api/grid/loadbalancer/edit',
                                           method='POST',
                                           params=params)
        except Exception as e:
            if "Update already pending" in str(e):
                raise LibcloudLBImmutableError(
                    "Balancer is immutable", GoGridLBDriver)

            raise LibcloudError(value='Exception: %s' % str(e), driver=self)

    def _members_to_params(self, members):
        """
        Helper method to convert list of :class:`Member` objects
        to GET params.

        """

        params = {}

        i = 0
        for member in members:
            params["realiplist.%s.ip" % i] = member.ip
            params["realiplist.%s.port" % i] = member.port
            i += 1

        return params

    def _to_balancers(self, object):
        return [self._to_balancer(el) for el in object["list"]]

    def _to_balancer(self, el):
        lb = LoadBalancer(id=el.get("id"),
                          name=el["name"],
                          state=self.LB_STATE_MAP.get(
                              el["state"]["name"], State.UNKNOWN),
                          ip=el["virtualip"]["ip"]["ip"],
                          port=el["virtualip"]["port"],
                          driver=self.connection.driver)
        return lb

    def _to_members(self, object, balancer=None):
        return [self._to_member(el, balancer) for el in object]

    def _to_member(self, el, balancer=None):
        member = Member(id=el["ip"]["id"],
                        ip=el["ip"]["ip"],
                        port=el["port"],
                        balancer=balancer)
        return member



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides generic deployment steps for machines post boot.
"""

from __future__ import with_statement

from typing import Union
from typing import Optional
from typing import List
from typing import IO
from typing import cast

import os
import re
import binascii

from libcloud.utils.py3 import basestring, PY3
from libcloud.compute.base import Node
from libcloud.compute.ssh import BaseSSHClient


class Deployment(object):
    """
    Base class for deployment tasks.
    """

    def run(self, node, client):
        # type: (Node, BaseSSHClient) -> Node
        """
        Runs this deployment task on node using the client provided.

        :type node: :class:`Node`
        :keyword node: Node to operate one

        :type client: :class:`BaseSSHClient`
        :keyword client: Connected SSH client to use.

        :return: :class:`Node`
        """
        raise NotImplementedError(
            'run not implemented for this deployment')

    def _get_string_value(self, argument_name, argument_value):
        if not isinstance(argument_value, basestring) and \
           not hasattr(argument_value, 'read'):
            raise TypeError('%s argument must be a string or a file-like '
                            'object' % (argument_name))

        if hasattr(argument_value, 'read'):
            argument_value = argument_value.read()

        return argument_value


class SSHKeyDeployment(Deployment):
    """
    Installs a public SSH Key onto a server.
    """

    def __init__(self, key):
        # type: (Union[str, IO]) -> None
        """
        :type key: ``str`` or :class:`File` object
        :keyword key: Contents of the public key write or a file object which
                      can be read.
        """
        self.key = self._get_string_value(argument_name='key',
                                          argument_value=key)

    def run(self, node, client):
        # type: (Node, BaseSSHClient) -> Node
        """
        Installs SSH key into ``.ssh/authorized_keys``

        See also :class:`Deployment.run`
        """
        client.put(".ssh/authorized_keys", contents=self.key, mode='a')
        return node

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        key = self.key[:100]
        return ("<SSHKeyDeployment key=%s...>" % (key))


class FileDeployment(Deployment):
    """
    Installs a file on the server.
    """

    def __init__(self, source, target):
        # type: (str, str) -> None
        """
        :type source: ``str``
        :keyword source: Local path of file to be installed

        :type target: ``str``
        :keyword target: Path to install file on node
        """
        self.source = source
        self.target = target

    def run(self, node, client):
        # type: (Node, BaseSSHClient) -> Node
        """
        Upload the file, retaining permissions.

        See also :class:`Deployment.run`
        """
        perms = int(oct(os.stat(self.source).st_mode)[4:], 8)

        with open(self.source, 'rb') as fp:
            client.putfo(path=self.target, chmod=perms,
                         fo=fp)
        return node

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return ("<FileDeployment source=%s, target=%s>" % (
            self.source, self.target))


class ScriptDeployment(Deployment):
    """
    Runs an arbitrary shell script on the server.

    This step works by first writing the content of the shell script (script
    argument) in a *.sh file on a remote server and then running that file.

    If you are running a non-shell script, make sure to put the appropriate
    shebang to the top of the script. You are also advised to do that even if
    you are running a plan shell script.
    """

    def __init__(self,
                 script,  # type: str
                 args=None,  # type: Optional[List[str]]
                 name=None,  # type: Optional[str]
                 delete=False,  # type bool
                 timeout=None  # type: Optional[float]
                 ):
        # type: (...) -> None
        """
        :type script: ``str``
        :keyword script: Contents of the script to run.

        :type args: ``list``
        :keyword args: Optional command line arguments which get passed to the
                       deployment script file.

        :type name: ``str``
        :keyword name: Name of the script to upload it as, if not specified,
                       a random name will be chosen.

        :type delete: ``bool``
        :keyword delete: Whether to delete the script on completion.

        :param timeout: Optional run timeout for this command.
        :type timeout: ``float``
        """
        script = self._get_string_value(argument_name='script',
                                        argument_value=script)

        self.script = script
        self.args = args or []
        self.stdout = None  # type: Optional[str]
        self.stderr = None  # type: Optional[str]
        self.exit_status = None  # type: Optional[int]
        self.delete = delete
        self.timeout = timeout
        self.name = name  # type: Optional[str]

        if self.name is None:
            # File is put under user's home directory
            # (~/libcloud_deployment_<random_string>.sh)
            random_string = ''  # type: Union[str, bytes]
            random_string = binascii.hexlify(os.urandom(4))
            random_string = cast(bytes, random_string)
            random_string = random_string.decode('ascii')
            self.name = 'libcloud_deployment_%s.sh' % (random_string)

    def run(self, node, client):
        # type: (Node, BaseSSHClient) -> Node
        """
        Uploads the shell script and then executes it.

        See also :class:`Deployment.run`
        """
        self.name = cast(str, self.name)
        file_path = client.put(path=self.name, chmod=int('755', 8),
                               contents=self.script)
        # Pre-pend cwd if user specified a relative path
        if self.name and (self.name[0] not in ['/', '\\'] and
                          not re.match(r"^\w\:.*$", file_path)):
            base_path = os.path.dirname(file_path)
            name = os.path.join(base_path, self.name)
        elif self.name and (self.name[0] == '\\' or
                            re.match(r"^\w\:.*$", file_path)):
            # Absolute Windows path
            name = file_path
        else:
            self.name = cast(str, self.name)
            name = self.name

        cmd = name

        if self.args:
            # Append arguments to the command
            cmd = '%s %s' % (name, ' '.join(self.args))
        else:
            cmd = name

        self.stdout, self.stderr, self.exit_status = \
            client.run(cmd, timeout=self.timeout)

        if self.delete:
            client.delete(self.name)

        return node

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        script = self.script[:15] + '...'
        exit_status = self.exit_status

        if exit_status is not None:
            stdout = self.stdout[:30] + '...'
            stderr = self.stderr[:30] + '...'
        else:
            exit_status = 'script didn\'t run yet'
            stdout = None
            stderr = None
        return ("<ScriptDeployment script=%s, exit_status=%s, stdout=%s, "
                "stderr=%s>" % (script, exit_status, stdout, stderr))


class ScriptFileDeployment(ScriptDeployment):
    """
    Runs an arbitrary shell script from a local file on the server. Same as
    ScriptDeployment, except that you can pass in a path to the file instead of
    the script content.
    """

    def __init__(self,
                 script_file,  # type: str
                 args=None,  # type: Optional[List[str]]
                 name=None,  # type: Optional[str]
                 delete=False,  # type bool
                 timeout=None  # type: Optional[float]
                 ):
        # type: (...) -> None
        """
        :type script_file: ``str``
        :keyword script_file: Path to a file containing the script to run.

        :type args: ``list``
        :keyword args: Optional command line arguments which get passed to the
                       deployment script file.


        :type name: ``str``
        :keyword name: Name of the script to upload it as, if not specified,
                       a random name will be chosen.

        :type delete: ``bool``
        :keyword delete: Whether to delete the script on completion.

        :param timeout: Optional run timeout for this command.
        :type timeout: ``float``
        """
        with open(script_file, 'rb') as fp:
            content = fp.read()  # type: Union[bytes, str]

        if PY3:
            content = cast(bytes, content)
            content = content.decode('utf-8')

        super(ScriptFileDeployment, self).__init__(script=content,
                                                   args=args,
                                                   name=name,
                                                   delete=delete,
                                                   timeout=timeout)


class MultiStepDeployment(Deployment):
    """
    Runs a chain of Deployment steps.
    """
    def __init__(self, add=None):
        # type: (Optional[Union[Deployment, List[Deployment]]]) -> None
        """
        :type add: ``list``
        :keyword add: Deployment steps to add.
        """
        self.steps = []  # type: list

        if add:
            self.add(add)

    def add(self, add):
        # type: (Union[Deployment, List[Deployment]]) -> None
        """
        Add a deployment to this chain.

        :type add: Single :class:`Deployment` or a ``list`` of
                   :class:`Deployment`
        :keyword add: Adds this deployment to the others already in this
                      object.
        """
        if add is not None:
            add = add if isinstance(add, (list, tuple)) else [add]
            self.steps.extend(add)

    def run(self, node, client):
        # type: (Node, BaseSSHClient) -> Node
        """
        Run each deployment that has been added.

        See also :class:`Deployment.run`
        """
        for s in self.steps:
            node = s.run(node, client)
        return node

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        steps = []

        for step in self.steps:
            steps.append(str(step))

        steps = ', '.join(steps)

        return ("<MultiStepDeployment steps=[%s]>" % (steps))



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.utils.py3 import httplib
from libcloud.dns.types import RecordType, ZoneDoesNotExistError
from libcloud.dns.types import RecordDoesNotExistError
from libcloud.dns.drivers.gandi import GandiDNSDriver
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.test.secrets import DNS_GANDI
from libcloud.test.common.test_gandi import BaseGandiMockHttp


class GandiTests(unittest.TestCase):

    def setUp(self):
        GandiDNSDriver.connectionCls.conn_class = GandiMockHttp
        GandiMockHttp.type = None
        self.driver = GandiDNSDriver(*DNS_GANDI)

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 10)
        self.assertTrue(RecordType.A in record_types)

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 5)

        zone = zones[0]
        self.assertEqual(zone.id, '47234')
        self.assertEqual(zone.type, 'master')
        self.assertEqual(zone.domain, 't.com')

    def test_list_records(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 4)

        record = records[1]
        self.assertEqual(record.name, 'www')
        self.assertEqual(record.id, 'A:www')
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, '208.111.35.173')

        record = records[3]
        self.assertEqual(record.name, '')
        self.assertEqual(record.id, 'MX:')
        self.assertEqual(record.type, RecordType.MX)
        self.assertEqual(record.data, 'aspmx.l.google.com')
        self.assertEqual(record.extra['priority'], 15)
        self.assertEqual(record.extra['ttl'], 86400)

    def test_get_zone(self):
        zone = self.driver.get_zone(zone_id='47234')
        self.assertEqual(zone.id, '47234')
        self.assertEqual(zone.type, 'master')
        self.assertEqual(zone.domain, 't.com')

    def test_get_record(self):
        record = self.driver.get_record(zone_id='47234',
                                        record_id='CNAME:t.com')
        self.assertEqual(record.name, 'wibble')
        self.assertEqual(record.type, RecordType.CNAME)
        self.assertEqual(record.data, 't.com')

    def test_list_records_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        GandiMockHttp.type = 'ZONE_DOES_NOT_EXIST'

        try:
            self.driver.list_records(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail('Exception was not thrown')

    def test_get_zone_does_not_exist(self):
        GandiMockHttp.type = 'ZONE_DOES_NOT_EXIST'

        try:
            self.driver.get_zone(zone_id='47234')
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, '47234')
        else:
            self.fail('Exception was not thrown')

    def test_get_record_zone_does_not_exist(self):
        GandiMockHttp.type = 'ZONE_DOES_NOT_EXIST'

        try:
            self.driver.get_record(zone_id='4444', record_id='CNAME:t.com')
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_get_record_record_does_not_exist(self):
        GandiMockHttp.type = 'RECORD_DOES_NOT_EXIST'

        try:
            self.driver.get_record(zone_id='47234',
                                   record_id='CNAME:t.com')
        except RecordDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_create_zone(self):
        zone = self.driver.create_zone(domain='t.com', type='master',
                                       ttl=None, extra=None)
        self.assertEqual(zone.id, '47234')
        self.assertEqual(zone.domain, 't.com')

    def test_update_zone(self):
        zone = self.driver.get_zone(zone_id='47234')
        zone = self.driver.update_zone(zone, domain='t.com')
        self.assertEqual(zone.id, '47234')
        self.assertEqual(zone.type, 'master')
        self.assertEqual(zone.domain, 't.com')

    def test_create_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(
            name='www', zone=zone,
            type=RecordType.A, data='127.0.0.1',
            extra={'ttl': 30}
        )

        self.assertEqual(record.id, 'A:www')
        self.assertEqual(record.name, 'www')
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, '127.0.0.1')

    def test_update_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[1]

        params = {
            'record': record,
            'name': 'www',
            'type': RecordType.A,
            'data': '127.0.0.1',
            'extra': {'ttl': 30}}
        updated_record = self.driver.update_record(**params)

        self.assertEqual(record.data, '208.111.35.173')

        self.assertEqual(updated_record.id, 'A:www')
        self.assertEqual(updated_record.name, 'www')
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, RecordType.A)
        self.assertEqual(updated_record.data, '127.0.0.1')

    def test_delete_zone(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        GandiMockHttp.type = 'ZONE_DOES_NOT_EXIST'

        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail('Exception was not thrown')

    def test_delete_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_does_not_exist(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        GandiMockHttp.type = 'RECORD_DOES_NOT_EXIST'
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record.id)
        else:
            self.fail('Exception was not thrown')


class GandiMockHttp(BaseGandiMockHttp):
    fixtures = DNSFileFixtures('gandi')

    def _xmlrpc__domain_zone_create(self, method, url, body, headers):
        body = self.fixtures.load('create_zone.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_update(self, method, url, body, headers):
        body = self.fixtures.load('get_zone.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_list(self, method, url, body, headers):
        body = self.fixtures.load('list_zones.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_list(self, method, url, body, headers):
        body = self.fixtures.load('list_records.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_add(self, method, url, body, headers):
        body = self.fixtures.load('create_record.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_delete(self, method, url, body, headers):
        body = self.fixtures.load('delete_zone.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_info(self, method, url, body, headers):
        body = self.fixtures.load('get_zone.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_delete(self, method, url, body, headers):
        body = self.fixtures.load('delete_record.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_update(self, method, url, body, headers):
        body = self.fixtures.load('create_record.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_version_new(self, method, url, body, headers):
        body = self.fixtures.load('new_version.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_version_set(self, method, url, body, headers):
        body = self.fixtures.load('new_version.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_list_ZONE_DOES_NOT_EXIST(self, method, url,
                                                             body, headers):
        body = self.fixtures.load('zone_doesnt_exist.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_info_ZONE_DOES_NOT_EXIST(self, method, url, body,
                                                      headers):
        body = self.fixtures.load('zone_doesnt_exist.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_list_ZONE_DOES_NOT_EXIST(self, method, url, body,
                                                      headers):
        body = self.fixtures.load('zone_doesnt_exist.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_delete_ZONE_DOES_NOT_EXIST(self, method, url,
                                                        body, headers):
        body = self.fixtures.load('zone_doesnt_exist.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_list_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('list_records_empty.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_info_RECORD_DOES_NOT_EXIST(self, method, url,
                                                        body, headers):
        body = self.fixtures.load('list_zones.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_record_delete_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('delete_record_doesnotexist.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_version_new_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('new_version.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__domain_zone_version_set_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('new_version.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement

import os
import sys
import platform
import shutil
import unittest
import time
import tempfile
import multiprocessing

from libcloud.common.types import LibcloudError
from libcloud.storage.base import Container
from libcloud.storage.base import Object
from libcloud.storage.types import ContainerDoesNotExistError
from libcloud.storage.types import ContainerAlreadyExistsError
from libcloud.storage.types import ContainerIsNotEmptyError
from libcloud.storage.types import InvalidContainerNameError
from libcloud.utils.files import exhaust_iterator

try:
    from libcloud.storage.drivers.local import LocalStorageDriver
    from libcloud.storage.drivers.local import LockLocalStorage
    import fasteners
except ImportError:
    print('fasteners library is not available, skipping local_storage tests...')
    LocalStorageDriver = None


class LocalTests(unittest.TestCase):
    driver_type = LocalStorageDriver

    @classmethod
    def create_driver(self):
        self.key = tempfile.mkdtemp()
        return self.driver_type(self.key, None)

    def setUp(self):
        self.driver = self.create_driver()

    def tearDown(self):
        shutil.rmtree(self.key)
        self.key = None

    def make_tmp_file(self, content=None):
        if not content:
            content = b'blah' * 1024
        _, tmppath = tempfile.mkstemp()
        with open(tmppath, 'wb') as fp:
            fp.write(content)
        return tmppath

    def remove_tmp_file(self, tmppath):
        try:
            os.unlink(tmppath)
        except Exception as e:
            msg = str(e)
            if 'being used by another process' in msg and platform.system().lower() == 'windows':
                return
            raise e

    @unittest.skipIf(platform.system().lower() == 'windows', 'Unsupported on Windows')
    def test_lock_local_storage(self):
        # 1. Acquire succeeds
        lock = LockLocalStorage("/tmp/a")
        with lock:
            self.assertTrue(True)

        # 2. Acquire fails because lock is already acquired
        lock = LockLocalStorage("/tmp/b", timeout=0.5)
        with lock:
            expected_msg = "Failed to acquire thread lock"
            self.assertRaisesRegex(LibcloudError, expected_msg, lock.__enter__)

        # 3. Multiprocessing scenario where IPC lock is involved
        def acquire_lock_in_subprocess(pid, success):
            # For first process acquire should succeed and for the second it should fail
            lock = LockLocalStorage("/tmp/c", timeout=0.5)

            if pid == 1:
                with lock:
                    # We use longer sleep when running tests in parallel to avoid
                    # failures related to slower process spawn
                    time.sleep(2.5)

                success.value = 1
            elif pid == 2:
                expected_msg = "Failed to acquire IPC lock"
                self.assertRaisesRegex(LibcloudError, expected_msg, lock.__enter__)
                success.value = 1
            else:
                raise ValueError("Invalid pid")

        success_1  = multiprocessing.Value('i', 0)
        success_2  = multiprocessing.Value('i', 0)

        p1 = multiprocessing.Process(target=acquire_lock_in_subprocess, args=(1, success_1,))
        p1.start()

        time.sleep(0.2)

        p2 = multiprocessing.Process(target=acquire_lock_in_subprocess, args=(2, success_2,))
        p2.start()

        p1.join()
        p2.join()

        self.assertEqual(bool(success_1.value), True, "Check didn't pass")
        self.assertEqual(bool(success_2.value), True, "Second check didn't pass")

    def test_list_containers_empty(self):
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 0)

    def test_containers_success(self):
        self.driver.create_container('test1')
        self.driver.create_container('test2')
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 2)

        container = containers[1]

        self.assertTrue('creation_time' in container.extra)
        self.assertTrue('modify_time' in container.extra)
        self.assertTrue('access_time' in container.extra)

        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

        objects = container.list_objects()
        self.assertEqual(len(objects), 0)

        for container in containers:
            self.driver.delete_container(container)

    def test_objects_success(self):
        tmppath = self.make_tmp_file()

        container = self.driver.create_container('test3')
        obj1 = container.upload_object(tmppath, 'object1')
        obj2 = container.upload_object(tmppath, 'path/object2')
        obj3 = container.upload_object(tmppath, 'path/to/object3')
        obj4 = container.upload_object(tmppath, 'path/to/object4.ext')
        with open(tmppath, 'rb') as tmpfile:
            obj5 = container.upload_object_via_stream(tmpfile, 'object5')

        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 5)

        prefix = os.path.join('path', 'invalid')
        objects = self.driver.list_container_objects(container=container,
                                                     prefix=prefix)
        self.assertEqual(len(objects), 0)

        prefix = os.path.join('path', 'to')
        objects = self.driver.list_container_objects(container=container,
                                                     prefix=prefix)
        self.assertEqual(len(objects), 2)

        for obj in objects:
            self.assertNotEqual(obj.hash, None)
            self.assertEqual(obj.size, 4096)
            self.assertEqual(obj.container.name, 'test3')
            self.assertTrue('creation_time' in obj.extra)
            self.assertTrue('modify_time' in obj.extra)
            self.assertTrue('access_time' in obj.extra)

        obj1.delete()
        obj2.delete()

        objects = container.list_objects()
        self.assertEqual(len(objects), 3)

        container.delete_object(obj3)
        container.delete_object(obj4)
        container.delete_object(obj5)

        objects = container.list_objects()
        self.assertEqual(len(objects), 0)

        container.delete()
        self.remove_tmp_file(tmppath)

    def test_get_container_doesnt_exist(self):
        try:
            self.driver.get_container(container_name='container1')
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_get_container_success(self):
        self.driver.create_container('test4')
        container = self.driver.get_container(container_name='test4')
        self.assertTrue(container.name, 'test4')
        container.delete()

    def test_get_object_container_doesnt_exist(self):
        try:
            self.driver.get_object(container_name='test-inexistent',
                                   object_name='test')
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_get_object_success(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container('test5')
        container.upload_object(tmppath, 'test')

        obj = self.driver.get_object(container_name='test5',
                                     object_name='test')

        self.assertEqual(obj.name, 'test')
        self.assertEqual(obj.container.name, 'test5')
        self.assertEqual(obj.size, 4096)
        self.assertNotEqual(obj.hash, None)
        self.assertTrue('creation_time' in obj.extra)
        self.assertTrue('modify_time' in obj.extra)
        self.assertTrue('access_time' in obj.extra)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)

    def test_create_container_invalid_name(self):
        try:
            self.driver.create_container(container_name='new/container')
        except InvalidContainerNameError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_create_container_already_exists(self):
        container = self.driver.create_container(
            container_name='new-container')
        try:
            self.driver.create_container(container_name='new-container')
        except ContainerAlreadyExistsError:
            pass
        else:
            self.fail('Exception was not thrown')

        # success
        self.driver.delete_container(container)

    def test_create_container_success(self):
        name = 'new_container'
        container = self.driver.create_container(container_name=name)
        self.assertEqual(container.name, name)
        self.driver.delete_container(container)

    def test_delete_container_doesnt_exist(self):
        container = Container(name='new_container', extra=None,
                              driver=self.driver)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_delete_container_not_empty(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        try:
            self.driver.delete_container(container=container)
        except ContainerIsNotEmptyError:
            pass
        else:
            self.fail('Exception was not thrown')

        # success
        obj.delete()
        self.remove_tmp_file(tmppath)
        self.assertTrue(self.driver.delete_container(container=container))

    def test_delete_container_not_found(self):
        container = Container(name='foo_bar_container', extra={},
                              driver=self.driver)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail('Container does not exist but an exception was not' +
                      'thrown')

    def test_delete_container_success(self):
        container = self.driver.create_container('test7')
        self.assertTrue(self.driver.delete_container(container=container))

    def test_download_object_success(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        destination_path = tmppath + '.temp'
        result = self.driver.download_object(obj=obj,
                                             destination_path=destination_path,
                                             overwrite_existing=False,
                                             delete_on_failure=True)

        self.assertTrue(result)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)
        os.unlink(destination_path)

    def test_download_object_and_overwrite(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        destination_path = tmppath + '.temp'
        result = self.driver.download_object(obj=obj,
                                             destination_path=destination_path,
                                             overwrite_existing=False,
                                             delete_on_failure=True)

        self.assertTrue(result)

        try:
            self.driver.download_object(obj=obj,
                                        destination_path=destination_path,
                                        overwrite_existing=False,
                                        delete_on_failure=True)
        except LibcloudError:
            pass
        else:
            self.fail('Exception was not thrown')

        result = self.driver.download_object(obj=obj,
                                             destination_path=destination_path,
                                             overwrite_existing=True,
                                             delete_on_failure=True)

        self.assertTrue(result)

        # success
        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)
        os.unlink(destination_path)

    def test_download_object_as_stream_success(self):
        tmppath = self.make_tmp_file()
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        stream = self.driver.download_object_as_stream(obj=obj,
                                                       chunk_size=1024)

        self.assertTrue(hasattr(stream, '__iter__'))

        data = b''.join(stream)
        self.assertTrue(len(data), 4096)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)

    def test_download_object_range_success(self):
        content = b'0123456789123456789'
        tmppath = self.make_tmp_file(content=content)
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        destination_path = tmppath + '.temp'

        # 1. Only start_bytes provided
        result = self.driver.download_object_range(obj=obj,
                                             destination_path=destination_path,
                                             start_bytes=4,
                                             overwrite_existing=True,
                                             delete_on_failure=True)
        self.assertTrue(result)

        with open(destination_path, 'rb') as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b'456789123456789')
        self.assertEqual(written_content, content[4:])

        # 2. start_bytes and end_bytes is provided
        result = self.driver.download_object_range(obj=obj,
                                             destination_path=destination_path,
                                             start_bytes=4,
                                             end_bytes=6,
                                             overwrite_existing=True,
                                             delete_on_failure=True)
        self.assertTrue(result)

        with open(destination_path, 'rb') as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b'45')
        self.assertEqual(written_content, content[4:6])

        result = self.driver.download_object_range(obj=obj,
                                             destination_path=destination_path,
                                             start_bytes=0,
                                             end_bytes=1,
                                             overwrite_existing=True,
                                             delete_on_failure=True)
        self.assertTrue(result)

        with open(destination_path, 'rb') as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b'0')
        self.assertEqual(written_content, content[0:1])

        result = self.driver.download_object_range(obj=obj,
                                             destination_path=destination_path,
                                             start_bytes=0,
                                             end_bytes=2,
                                             overwrite_existing=True,
                                             delete_on_failure=True)
        self.assertTrue(result)

        with open(destination_path, 'rb') as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b'01')
        self.assertEqual(written_content, content[0:2])

        result = self.driver.download_object_range(obj=obj,
                                             destination_path=destination_path,
                                             start_bytes=0,
                                             end_bytes=len(content),
                                             overwrite_existing=True,
                                             delete_on_failure=True)
        self.assertTrue(result)

        with open(destination_path, 'rb') as fp:
            written_content = fp.read()

        self.assertEqual(written_content, b'0123456789123456789')
        self.assertEqual(written_content, content[0:len(content)])

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)
        os.unlink(destination_path)

    def test_download_object_range_as_stream_success(self):
        content = b'0123456789123456789'
        tmppath = self.make_tmp_file(content=content)
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        # 1. Only start_bytes provided
        stream = self.driver.download_object_range_as_stream(obj=obj,
                                                             start_bytes=4,
                                                             chunk_size=1024)
        written_content = b''.join(stream)

        self.assertEqual(written_content, b'456789123456789')
        self.assertEqual(written_content, content[4:])

        # 2. start_bytes and end_bytes is provided
        stream = self.driver.download_object_range_as_stream(obj=obj,
                                                             start_bytes=4,
                                                             end_bytes=7,
                                                             chunk_size=1024)
        written_content = b''.join(stream)

        self.assertEqual(written_content, b'456')
        self.assertEqual(written_content, content[4:7])

        stream = self.driver.download_object_range_as_stream(obj=obj,
                                                             start_bytes=0,
                                                             end_bytes=1,
                                                             chunk_size=1024)
        written_content = b''.join(stream)

        self.assertEqual(written_content, b'0')
        self.assertEqual(written_content, content[0:1])

        stream = self.driver.download_object_range_as_stream(obj=obj,
                                                             start_bytes=1,
                                                             end_bytes=3,
                                                             chunk_size=1024)
        written_content = b''.join(stream)

        self.assertEqual(written_content, b'12')
        self.assertEqual(written_content, content[1:3])

        stream = self.driver.download_object_range_as_stream(obj=obj,
                                                             start_bytes=0,
                                                             end_bytes=len(content),
                                                             chunk_size=1024)
        written_content = b''.join(stream)

        self.assertEqual(written_content, b'0123456789123456789')
        self.assertEqual(written_content, content[0:len(content)])

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)

    def test_download_object_range_invalid_values(self):
        obj = Object('a', 500, '', {}, {}, None, None)
        tmppath = self.make_tmp_file(content='')

        expected_msg = 'start_bytes must be greater than 0'
        self.assertRaisesRegex(ValueError, expected_msg,
            self.driver.download_object_range, obj=obj,
            destination_path=tmppath,
            start_bytes=-1)

        expected_msg = 'start_bytes must be smaller than end_bytes'
        self.assertRaisesRegex(ValueError, expected_msg,
            self.driver.download_object_range, obj=obj,
            destination_path=tmppath,
            start_bytes=5,
            end_bytes=4)

        expected_msg = 'start_bytes and end_bytes can\'t be the same'
        self.assertRaisesRegex(ValueError, expected_msg,
            self.driver.download_object_range, obj=obj,
            destination_path=tmppath,
            start_bytes=5,
            end_bytes=5)

    def test_download_object_range_as_stream_invalid_values(self):
        content = b'0123456789123456789'
        tmppath = self.make_tmp_file(content=content)
        container = self.driver.create_container('test6')
        obj = container.upload_object(tmppath, 'test')

        expected_msg = 'start_bytes must be greater than 0'
        stream = self.driver.download_object_range_as_stream(
            obj=obj,
            start_bytes=-1,
            end_bytes=None,
            chunk_size=1024)
        self.assertRaisesRegex(ValueError, expected_msg, exhaust_iterator,
                               stream)

        expected_msg = 'start_bytes must be smaller than end_bytes'
        stream = self.driver.download_object_range_as_stream(
            obj=obj,
            start_bytes=5,
            end_bytes=4,
            chunk_size=1024)
        self.assertRaisesRegex(ValueError, expected_msg, exhaust_iterator,
                               stream)

        expected_msg = 'end_bytes is larger than file size'
        stream = self.driver.download_object_range_as_stream(
            obj=obj,
            start_bytes=5,
            end_bytes=len(content) + 1,
            chunk_size=1024)

        expected_msg = 'start_bytes and end_bytes can\'t be the same'
        stream = self.driver.download_object_range_as_stream(
            obj=obj,
            start_bytes=5,
            end_bytes=5,
            chunk_size=1024)

        obj.delete()
        container.delete()
        self.remove_tmp_file(tmppath)


if not LocalStorageDriver:
    class LocalTests(unittest.TestCase):  # NOQA
        pass


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import sys

from libcloud.compute.base import Node
from libcloud.compute.drivers.onapp import OnAppNodeDriver
from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.test.secrets import ONAPP_PARAMS
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.utils.py3 import httplib


class OnAppNodeTestCase(LibcloudTestCase):
    driver_klass = OnAppNodeDriver

    def setUp(self):
        self.driver_klass.connectionCls.conn_class = OnAppMockHttp

        self.driver = OnAppNodeDriver(*ONAPP_PARAMS)

    def test_create_node(self):
        node = self.driver.create_node(
            name='onapp-new-fred',
            ex_memory=512,
            ex_cpus=4,
            ex_cpu_shares=4,
            ex_hostname='onapp-new-fred',
            ex_template_id='template_id',
            ex_primary_disk_size=100,
            ex_swap_disk_size=1,
            ex_required_virtual_machine_build=0,
            ex_required_ip_address_assignment=0
        )

        extra = node.extra

        self.assertEqual('onapp-new-fred', node.name)
        self.assertEqual('456789', node.id)
        self.assertEqual('456789', node.id)

        self.assertEqual('delivered', node.state)

        self.assertEqual(True, extra['booted'])
        self.assertEqual('passwd', extra['initial_root_password'])
        self.assertEqual('8.8.8.8', extra['local_remote_access_ip_address'])

        self.assertEqual(['192.168.15.73'], node.private_ips)
        self.assertEqual([], node.public_ips)

    def test_destroy_node(self):
        node = Node('identABC', 'testnode',
                    ['123.123.123.123'], [],
                    {'state': 'test', 'template_id': 88}, None)
        res = self.driver.destroy_node(node=node)
        self.assertTrue(res)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        extra = nodes[0].extra
        private_ips = nodes[0].private_ips

        self.assertEqual(1, len(nodes))
        self.assertEqual('onapp-fred', nodes[0].name)
        self.assertEqual('123456', nodes[0].id)

        self.assertEqual(True, extra['booted'])
        self.assertEqual('passwd', extra['initial_root_password'])
        self.assertEqual('9.9.9.9', extra['local_remote_access_ip_address'])

        self.assertEqual(1, len(private_ips))
        self.assertEqual('192.168.15.72', private_ips[0])

    def test_list_images(self):
        images = self.driver.list_images()
        extra = images[0].extra

        self.assertEqual(1, len(images))
        self.assertEqual('CentOS 5.11 x64', images[0].name)
        self.assertEqual('123456', images[0].id)

        self.assertEqual(True, extra['allowed_swap'])
        self.assertEqual(256, extra['min_memory_size'])
        self.assertEqual('rhel', extra['distribution'])

    def test_list_key_pairs(self):
        keys = self.driver.list_key_pairs()
        self.assertEqual(2, len(keys))
        self.assertEqual(1, keys[0].name)
        self.assertIsNotNone(keys[0].public_key)
        self.assertIsNotNone(keys[1].public_key)

    def test_get_key_pair(self):
        key = self.driver.get_key_pair(1)
        self.assertEqual(1, key.name)
        self.assertIsNotNone(key.public_key)

    def test_import_key_pair_from_string(self):
        key = self.driver.import_key_pair_from_string(
            'name',
            'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC8uuUq')
        self.assertEqual(3, key.name)
        self.assertEqual(
            'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC8uuUq',
            key.public_key)

    def test_delete_key_pair(self):
        key = self.driver.get_key_pair(1)
        response = self.driver.delete_key_pair(key)
        self.assertTrue(response)


class OnAppMockHttp(MockHttp):
    fixtures = ComputeFileFixtures('onapp')

    def _virtual_machines_json(self, method, url, body, headers):
        if method == 'GET':
            body = self.fixtures.load('list_nodes.json')
        else:
            body = self.fixtures.load('create_node.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _virtual_machines_identABC_json(self, method, url, body, headers):
        return (
            httplib.NO_CONTENT,
            '',
            {},
            httplib.responses[httplib.NO_CONTENT]
        )

    def _templates_json(self, method, url, body, headers):
        body = self.fixtures.load('list_images.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _profile_json(self, method, url, body, headers):
        body = self.fixtures.load('profile.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _users_123_ssh_keys_json(self, method, url, body, headers):
        if method == 'GET':
            body = self.fixtures.load('list_key_pairs.json')
        else:
            body = self.fixtures.load('import_key_pair.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _users_123_ssh_keys_1_json(self, method, url, body, headers):
        body = self.fixtures.load('get_key_pair.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _settings_ssh_keys_1_json(self, method, url, body, headers):
        return (httplib.NO_CONTENT, '', {},
                httplib.responses[httplib.NO_CONTENT])


if __name__ == '__main__':
    sys.exit(unittest.main())





# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'GCEBackupDriver'
]

from libcloud.utils.iso8601 import parse_date
from libcloud.backup.base import BackupDriver, BackupTargetRecoveryPoint,\
    BackupTargetJob, BackupTarget
from libcloud.backup.types import BackupTargetType, BackupTargetJobStatusType

from libcloud.common.google import GoogleResponse, GoogleBaseConnection

API_VERSION = 'v1'
DEFAULT_TASK_COMPLETION_TIMEOUT = 180


class GCEResponse(GoogleResponse):
    pass


class GCEConnection(GoogleBaseConnection):
    """
    Connection class for the GCE driver.

    GCEConnection extends :class:`google.GoogleBaseConnection` for 2 reasons:
      1. modify request_path for GCE URI.
      2. Implement gce_params functionality described below.

    If the parameter gce_params is set to a dict prior to calling request(),
    the URL parameters will be updated to include those key/values FOR A
    SINGLE REQUEST. If the response contains a nextPageToken,
    gce_params['pageToken'] will be set to its value. This can be used to
    implement paging in list:

    >>> params, more_results = {'maxResults': 2}, True
    >>> while more_results:
    ...     driver.connection.gce_params=params
    ...     driver.ex_list_urlmaps()
    ...     more_results = 'pageToken' in params
    ...
    [<GCEUrlMap id="..." name="cli-map">, <GCEUrlMap id="..." name="lc-map">]
    [<GCEUrlMap id="..." name="web-map">]
    """
    host = 'www.googleapis.com'
    responseCls = GCEResponse

    def __init__(self, user_id, key, secure, auth_type=None,
                 credential_file=None, project=None, **kwargs):
        super(GCEConnection, self).__init__(user_id, key, secure=secure,
                                            auth_type=auth_type,
                                            credential_file=credential_file,
                                            **kwargs)
        self.request_path = '/compute/%s/projects/%s' % (API_VERSION,
                                                         project)
        self.gce_params = {}

    def pre_connect_hook(self, params, headers):
        """
        Update URL parameters with values from self.gce_params.

        @inherits: :class:`GoogleBaseConnection.pre_connect_hook`
        """
        params, headers = super(GCEConnection, self).pre_connect_hook(params,
                                                                      headers)
        if self.gce_params:
            params.update(self.gce_params)
        return params, headers

    def request(self, *args, **kwargs):
        """
        Perform request then do GCE-specific processing of URL params.

        @inherits: :class:`GoogleBaseConnection.request`
        """
        response = super(GCEConnection, self).request(*args, **kwargs)

        # If gce_params has been set, then update the pageToken with the
        # nextPageToken so it can be used in the next request.
        if self.gce_params:
            if 'nextPageToken' in response.object:
                self.gce_params['pageToken'] = response.object['nextPageToken']
            elif 'pageToken' in self.gce_params:
                del self.gce_params['pageToken']
            self.gce_params = None

        return response


class GCEBackupDriver(BackupDriver):
    name = 'Google Compute Engine Backup Driver'
    website = 'http://cloud.google.com/'
    connectionCls = GCEConnection

    def __init__(self, user_id, key=None, project=None,
                 auth_type=None, scopes=None, credential_file=None, **kwargs):
        """
        :param  user_id: The email address (for service accounts) or Client ID
                         (for installed apps) to be used for authentication.
        :type   user_id: ``str``

        :param  key: The RSA Key (for service accounts) or file path containing
                     key or Client Secret (for installed apps) to be used for
                     authentication.
        :type   key: ``str``

        :keyword  project: Your GCE project name. (required)
        :type     project: ``str``

        :keyword  auth_type: Accepted values are "SA" or "IA" or "GCE"
                             ("Service Account" or "Installed Application" or
                             "GCE" if libcloud is being used on a GCE instance
                             with service account enabled).
                             If not supplied, auth_type will be guessed based
                             on value of user_id or if the code is being
                             executed in a GCE instance.
        :type     auth_type: ``str``

        :keyword  scopes: List of authorization URLs. Default is empty and
                          grants read/write to Compute, Storage, DNS.
        :type     scopes: ``list``

        :keyword  credential_file: Path to file for caching authentication
                                   information used by GCEConnection.
        :type     credential_file: ``str``
        """
        if not project:
            raise ValueError('Project name must be specified using '
                             '"project" keyword.')

        self.auth_type = auth_type
        self.project = project
        self.scopes = scopes
        self.credential_file = credential_file or \
            '~/.gce_libcloud_auth' + '.' + self.project

        super(GCEBackupDriver, self).__init__(user_id, key, **kwargs)

        # Cache Zone and Region information to reduce API calls and
        # increase speed
        self.base_path = '/compute/%s/projects/%s' % (API_VERSION,
                                                      self.project)

    def get_supported_target_types(self):
        """
        Get a list of backup target types this driver supports

        :return: ``list`` of :class:``BackupTargetType``
        """
        return [BackupTargetType.VOLUME]

    def list_targets(self):
        """
        List all backuptargets

        :rtype: ``list`` of :class:`BackupTarget`
        """
        raise NotImplementedError(
            'list_targets not implemented for this driver')

    def create_target(self, name, address,
                      type=BackupTargetType.VOLUME, extra=None):
        """
        Creates a new backup target

        :param name: Name of the target
        :type name: ``str``

        :param address: The volume ID.
        :type address: ``str``

        :param type: Backup target type (Physical, Virtual, ...).
        :type type: :class:`BackupTargetType`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        # Does nothing since any volume can be snapped at anytime.
        return self.ex_get_target_by_source(address)

    def create_target_from_node(self, node, type=BackupTargetType.VIRTUAL,
                                extra=None):
        """
        Creates a new backup target from an existing node

        :param node: The Node to backup
        :type  node: ``Node``

        :param type: Backup target type (Physical, Virtual, ...).
        :type type: :class:`BackupTargetType`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        # Get the first persistent disk
        disks = node.extra['disks']
        if disks is not None:
            return self.create_target(
                name=node.name,
                address=disks[0]['source'],
                type=BackupTargetType.VOLUME,
                extra=None)
        else:
            raise RuntimeError("Node does not have any block devices")

    def create_target_from_container(self, container,
                                     type=BackupTargetType.OBJECT,
                                     extra=None):
        """
        Creates a new backup target from an existing storage container

        :param node: The Container to backup
        :type  node: ``Container``

        :param type: Backup target type (Physical, Virtual, ...).
        :type type: :class:`BackupTargetType`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        raise NotImplementedError(
            'create_target_from_container not implemented for this driver')

    def update_target(self, target, name, address, extra):
        """
        Update the properties of a backup target

        :param target: Backup target to update
        :type  target: Instance of :class:`BackupTarget`

        :param name: Name of the target
        :type name: ``str``

        :param address: Hostname, FQDN, IP, file path etc.
        :type address: ``str``

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTarget`
        """
        # Does nothing since any volume can be snapped at anytime.
        return self.ex_get_target_by_source(address)

    def delete_target(self, target):
        """
        Delete a backup target

        :param target: Backup target to delete
        :type  target: Instance of :class:`BackupTarget`
        """
        raise NotImplementedError(
            'delete_target not implemented for this driver')

    def list_recovery_points(self, target, start_date=None, end_date=None):
        """
        List the recovery points available for a target

        :param target: Backup target to delete
        :type  target: Instance of :class:`BackupTarget`

        :param start_date: The start date to show jobs between (optional)
        :type  start_date: :class:`datetime.datetime`

        :param end_date: The end date to show jobs between (optional)
        :type  end_date: :class:`datetime.datetime``

        :rtype: ``list`` of :class:`BackupTargetRecoveryPoint`
        """
        request = '/global/snapshots'
        response = self.connection.request(request, method='GET').object
        return self._to_recovery_points(response, target)

    def recover_target(self, target, recovery_point, path=None):
        """
        Recover a backup target to a recovery point

        :param target: Backup target to delete
        :type  target: Instance of :class:`BackupTarget`

        :param recovery_point: Backup target with the backup data
        :type  recovery_point: Instance of :class:`BackupTarget`

        :param path: The part of the recovery point to recover (optional)
        :type  path: ``str``

        :rtype: Instance of :class:`BackupTargetJob`
        """
        raise NotImplementedError(
            'recover_target not implemented for this driver')

    def recover_target_out_of_place(self, target, recovery_point,
                                    recovery_target, path=None):
        """
        Recover a backup target to a recovery point out-of-place

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param recovery_point: Backup target with the backup data
        :type  recovery_point: Instance of :class:`BackupTarget`

        :param recovery_target: Backup target with to recover the data to
        :type  recovery_target: Instance of :class:`BackupTarget`

        :param path: The part of the recovery point to recover (optional)
        :type  path: ``str``

        :rtype: Instance of :class:`BackupTargetJob`
        """
        raise NotImplementedError(
            'recover_target_out_of_place not implemented for this driver')

    def get_target_job(self, target, id):
        """
        Get a specific backup job by ID

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param id: Backup target with the backup data
        :type  id: Instance of :class:`BackupTarget`

        :rtype: :class:`BackupTargetJob`
        """
        jobs = self.list_target_jobs(target)
        return list(filter(lambda x: x.id == id, jobs))[0]

    def list_target_jobs(self, target):
        """
        List the backup jobs on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :rtype: ``list`` of :class:`BackupTargetJob`
        """
        return []

    def create_target_job(self, target, extra=None):
        """
        Create a new backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``

        :rtype: Instance of :class:`BackupTargetJob`
        """
        name = target.name
        request = '/zones/%s/disks/%s/createSnapshot' % (
            target.extra['zone'].name, target.name)
        snapshot_data = {
            'source': target.extra['source']
        }
        self.connection.async_request(request, method='POST',
                                      data=snapshot_data)
        return self._to_job(self.ex_get_snapshot(name), target)

    def resume_target_job(self, target, job):
        """
        Resume a suspended backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param job: Backup target job to resume
        :type  job: Instance of :class:`BackupTargetJob`

        :rtype: ``bool``
        """
        raise NotImplementedError(
            'resume_target_job not supported for this driver')

    def suspend_target_job(self, target, job):
        """
        Suspend a running backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param job: Backup target job to suspend
        :type  job: Instance of :class:`BackupTargetJob`

        :rtype: ``bool``
        """
        raise NotImplementedError(
            'suspend_target_job not supported for this driver')

    def cancel_target_job(self, target, job):
        """
        Cancel a backup job on a target

        :param target: Backup target with the backup data
        :type  target: Instance of :class:`BackupTarget`

        :param job: Backup target job to cancel
        :type  job: Instance of :class:`BackupTargetJob`

        :rtype: ``bool``
        """
        raise NotImplementedError(
            'cancel_target_job not supported for this driver')

    def _to_recovery_points(self, data, target):
        return [self._to_recovery_point(item, target)
                for item in data.items]

    def _to_recovery_point(self, item, target):
        id = item.id
        date = parse_date(item.creationTimestamp)
        point = BackupTargetRecoveryPoint(
            id=id,
            date=date,
            target=target,
            driver=self.connection.driver,
            extra={
                'snapshot-id': id,
            },
        )
        return point

    def _to_jobs(self, data, target):
        return [self._to_job(item, target)
                for item in data.items]

    def _to_job(self, item, target):
        id = item.id
        job = BackupTargetJob(
            id=id,
            status=BackupTargetJobStatusType.PENDING,
            progress=0,
            target=target,
            driver=self.connection.driver,
            extra={
            },
        )
        return job

    def ex_get_snapshot(self, name):
        request = '/global/snapshots/%s' % (name)
        response = self.connection.request(request, method='GET').object
        return response

    def ex_get_target_by_source(self, source):
        return BackupTarget(
            id=source,
            name=source,
            address=source,
            type=BackupTargetType.VOLUME,
            driver=self.connection.driver,
            extra={
                "source": source
            }
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
VPS.net driver
"""
import base64

try:
    import simplejson as json
except ImportError:
    import json

from libcloud.utils.py3 import b

from libcloud.common.base import ConnectionUserAndKey, JsonResponse
from libcloud.common.types import InvalidCredsError, MalformedResponseError
from libcloud.compute.providers import Provider
from libcloud.compute.types import NodeState
from libcloud.compute.base import Node, NodeDriver
from libcloud.compute.base import NodeSize, NodeImage, NodeLocation

API_HOST = 'api.vps.net'
API_VERSION = 'api10json'

RAM_PER_NODE = 256
DISK_PER_NODE = 10
BANDWIDTH_PER_NODE = 250


class VPSNetResponse(JsonResponse):
    def parse_body(self):
        try:
            return super(VPSNetResponse, self).parse_body()
        except MalformedResponseError:
            return self.body

    def success(self):
        # vps.net wrongly uses 406 for invalid auth creds
        if self.status == 406 or self.status == 403:
            raise InvalidCredsError()
        return True

    def parse_error(self):
        try:
            errors = super(VPSNetResponse, self).parse_body()['errors'][0]
        except MalformedResponseError:
            return self.body
        else:
            return "\n".join(errors)


class VPSNetConnection(ConnectionUserAndKey):
    """
    Connection class for the VPS.net driver
    """

    host = API_HOST
    responseCls = VPSNetResponse

    allow_insecure = False

    def add_default_headers(self, headers):
        user_b64 = base64.b64encode(b('%s:%s' % (self.user_id, self.key)))
        headers['Authorization'] = 'Basic %s' % (user_b64.decode('utf-8'))
        return headers


class VPSNetNodeDriver(NodeDriver):
    """
    VPS.net node driver
    """

    type = Provider.VPSNET
    api_name = 'vps_net'
    name = "vps.net"
    website = 'http://vps.net/'
    connectionCls = VPSNetConnection

    def _to_node(self, vm):
        if vm['running']:
            state = NodeState.RUNNING
        else:
            state = NodeState.PENDING

        n = Node(id=vm['id'],
                 name=vm['label'],
                 state=state,
                 public_ips=[vm.get('primary_ip_address', None)],
                 private_ips=[],
                 extra={'slices_count': vm['slices_count']},
                 # Number of nodes consumed by VM
                 driver=self.connection.driver)
        return n

    def _to_image(self, image, cloud):
        image = NodeImage(id=image['id'],
                          name="%s: %s" % (cloud, image['label']),
                          driver=self.connection.driver)

        return image

    def _to_size(self, num):
        size = NodeSize(id=num,
                        name="%d Node" % (num,),
                        ram=RAM_PER_NODE * num,
                        disk=DISK_PER_NODE,
                        bandwidth=BANDWIDTH_PER_NODE * num,
                        price=self._get_price_per_node(num) * num,
                        driver=self.connection.driver)
        return size

    def _get_price_per_node(self, num):
        single_node_price = self._get_size_price(size_id='1')
        return num * single_node_price

    def create_node(self, name, image, size, ex_backups_enabled=False,
                    ex_fqdn=None):
        """Create a new VPS.net node

        @inherits: :class:`NodeDriver.create_node`

        :keyword    ex_backups_enabled: Enable automatic backups
        :type       ex_backups_enabled: ``bool``

        :keyword    ex_fqdn:   Fully Qualified domain of the node
        :type       ex_fqdn:   ``str``
        """
        ex_backups_enabled = 1 if ex_backups_enabled else 0
        headers = {'Content-Type': 'application/json'}
        request = {'virtual_machine':
                   {'label': name,
                    'fqdn': ex_fqdn or '',
                    'system_template_id': image.id,
                    'backups_enabled': ex_backups_enabled,
                    'slices_required': size.id}}

        res = self.connection.request('/virtual_machines.%s' % (API_VERSION,),
                                      data=json.dumps(request),
                                      headers=headers,
                                      method='POST')
        node = self._to_node(res.object['virtual_machine'])
        return node

    def reboot_node(self, node):
        res = self.connection.request(
            '/virtual_machines/%s/%s.%s' % (node.id,
                                            'reboot',
                                            API_VERSION),
            method="POST")
        node = self._to_node(res.object['virtual_machine'])
        return True

    def list_sizes(self, location=None):
        res = self.connection.request('/nodes.%s' % (API_VERSION,))
        available_nodes = len([size for size in res.object
                               if size['slice']['virtual_machine_id']])
        sizes = [self._to_size(i) for i in range(1, available_nodes + 1)]
        return sizes

    def destroy_node(self, node):
        res = self.connection.request('/virtual_machines/%s.%s'
                                      % (node.id, API_VERSION),
                                      method='DELETE')
        return res.status == 200

    def list_nodes(self):
        res = self.connection.request('/virtual_machines.%s' % (API_VERSION,))
        return [self._to_node(i['virtual_machine']) for i in res.object]

    def list_images(self, location=None):
        res = self.connection.request('/available_clouds.%s' % (API_VERSION,))

        images = []
        for cloud in res.object:
            label = cloud['cloud']['label']
            templates = cloud['cloud']['system_templates']
            images.extend([self._to_image(image, label)
                           for image in templates])

        return images

    def list_locations(self):
        return [NodeLocation(0, "VPS.net Western US", 'US', self)]
